from typing import List, Dict

import numpy as np

from line_profiler import profile

from skanym.structures.character.skeleton import Skeleton
from skanym.structures.animation.animation import Animation
from skanym.structures.data.vectorizedArray import VectorizedArray
from skanym.animators.baseFkAnimator import BaseFkAnimator, PlayMode


class LocalFkAnimator(BaseFkAnimator):
    """ """
    def __init__(self, skeleton: Skeleton, animations: List[Animation], play_mode: PlayMode = PlayMode.LOOP_ALL):        
        self.vec_bind_transform_matrices = None
        super().__init__(skeleton, animations, play_mode)     

    # @profile
    def _initialize(self):
        super()._initialize()       
        self.vec_bind_transform_matrices = self.skeleton.as_bind_transform_vector()        
        self.vec_bind_transform_matrices = self.vec_bind_transform_matrices.tile((self.nb_anims, 1, 1))

    # @profile
    def _execute(self):        
        lst_output = super()._execute()
        vec_delta_transform_matrices = lst_output[0]

        arr_local_transform_matrices = self.vec_bind_transform_matrices.get() @ vec_delta_transform_matrices.get()

        vec_local_transform_matrices = VectorizedArray(arr_local_transform_matrices, labels=["local_transform"])

        lst_output.append(vec_local_transform_matrices)
        return lst_output



import struct
import warnings
from typing import List, Tuple
from pathlib import Path

import numpy as np

from skanym.utils.iFileLoader import IFileLoader
from skanym.core.model.skeleton import Skeleton
from skanym.core.model.joint import Joint
from skanym.core.animate.animation import Animation
from skanym.core.animate.curve.animationCurve import AnimationCurve
from skanym.core.animate.curve.positionCurve import PositionCurve
from skanym.core.animate.curve.quaternionCurve import QuaternionCurve
from skanym.core.animate.curve.key import Key
from skanym.core.math.transform import Transform


class GlbLoader(IFileLoader):
    def __init__(self, path: Path):
        super().__init__(path)
        try:
            import bpy
        except ImportError:
            warnings.warn("bpy module not found. Blender is not installed.")
            return
        
        try:
            from pygltflib import GLTF2
        except ImportError:
            warnings.warn("pygltflib module not found.")
            return
        
        self._GLTF2 = GLTF2
        
    
    def load_skeleton(self) -> Skeleton:
        return self._load_gltf()[0]
    
    def load_animation(self) -> Animation:
        return self._load_gltf()[1]
    
    def load_animations(self) -> List[Animation]:
        return NotImplementedError
    
    def _load_gltf(self) -> Tuple[Skeleton, Animation]:
        """TODO rewrite
        Creates an Animator object from a glTF 2.0 file.
        To create an Animator, we need the following piece of data:
            1. Skeleton Hierarchy
            2. Animator
                1. Joint Animations
                    1. Animation Curves

        glTF provides us two interesting lists: skins and animations
        We'll retrieve the Armature hierarchy from the skin list
        Then, we'll get the animation curves from the animation list

        We have to be careful Because the animation data contains
        final pos/rot and not the differences

        We can directly get the local_bind_transform from the metadata
        We need the accessor-buffer chain to get the time + TRS curves
        We are only interested in the translation and rotation though

        Same assumptions: one skin, one anim. No mesh yet.
        """
        
        def _array_from_accessor_id(accessor_i):
            accessor = gltf.accessors[accessor_i]
            buffer_view = gltf.bufferViews[accessor.bufferView]
            buffer = gltf.buffers[buffer_view.buffer]

            data = gltf.get_data_from_buffer_uri(buffer.uri)
            struct_size = buffer_view.byteLength // accessor.count
            output = []
            for i in range(accessor.count):
                index = buffer_view.byteOffset + accessor.byteOffset + i * struct_size
                d = data[index : index + struct_size]
                v = struct.unpack("<" + (struct_size // 4) * "f", d)
                if len(v) == 1:
                    v = v[0]
                output.append(v)

            return output

        gltf = self._GLTF2().load(self._path)

        skeleton_nodes = {}
        skeleton_root = None

        curves = {"translation": {}, "rotation": {}, "scale": {}, "weights": {}}
        anim_duration = 1.0

        # 1. Create a list of Joints from the file.
        for joint_i in gltf.skins[0].joints:
            skeleton_node = gltf.nodes[joint_i]

            joint_name = skeleton_node.name
            pos = skeleton_node.translation
            rot = skeleton_node.rotation

            if not pos:
                pos = [0.0, 0.0, 0.0]
            if not rot:
                rot = [0.0, 0.0, 0.0, 1.0]

            orient = np.quaternion(rot[3], rot[0], rot[1], rot[2])
            xform = Transform(pos, orient.normalized())
            joint = Joint(name=joint_name, local_bind_transform=xform)

            skeleton_nodes[joint_i] = joint
            curves["translation"][joint_i] = None
            curves["rotation"][joint_i] = None

        # 2. Create the skeleton hierarchy
        for joint_i in skeleton_nodes:
            skeleton_node = gltf.nodes[joint_i]
            children = [skeleton_nodes[child] for child in skeleton_node.children]
            skeleton_nodes[joint_i].add_children(children)

        armature_node = gltf.nodes[gltf.scenes[gltf.scene].nodes[0]]
        skeleton_root = skeleton_nodes[armature_node.children[0]]
        skeleton_name = armature_node.name
        
        output_skeleton = Skeleton(
            name=skeleton_name,
            root=skeleton_root,
        )

        # # 3. Create the animation curves from the file
        animation = gltf.animations[0]

        output_animation = Animation(
            name=self._path.stem,
            duration=0.0,
            shift=0.0,
            position_curves=[],
            rotation_curves=[],
        )
        
        for channel in animation.channels:
            target = channel.target
            joint_i = target.node

            sampler = animation.samplers[channel.sampler]
            key_times = _array_from_accessor_id(sampler.input)
            key_data = _array_from_accessor_id(sampler.output)
            keys = []

            key_len = len(key_times)
            duration = 1.0
            if key_len > 0:
                duration = key_times[-1]

                if key_len - 1 > anim_duration:
                    anim_duration = float(key_len - 1)

            for i in range(len(key_times)):
                t = float(int((key_times[i] / duration) * (key_len - 1)))
                v = key_data[i]
                if target.path == "translation":
                    p = skeleton_nodes[joint_i].local_bind_transform.getPosition()
                    if joint_i == armature_node.children[0]:
                        # Axis conversion for root joint
                        skeleton_nodes[joint_i].local_bind_transform.position = [p[0], -p[2], p[1]]
                        p = [0.0, 0.0, 0.0]
                        v = [v[0], -v[2], v[1]] 
                    keys.append(Key(t, np.array([v[0] - p[0], v[1] - p[1], v[2] - p[2]])))               
                
                elif target.path == "rotation":
                    q = skeleton_nodes[joint_i].local_bind_transform.getRotation()
                    Q1 = np.quaternion(v[3], v[0], v[1], v[2])
                    Q1 = Q1.normalized()
                    if joint_i == armature_node.children[0]:
                        keys.append(Key(t, Q1))
                    else:
                        keys.append(Key(t, q.conjugate() * Q1))
                else:
                    break

            for key in keys:
                key.time /= key_len - 1

            # if target.path == "translation":
            #     for key in keys:
            #         print(key.time, key.value)

            if target.path == "translation":
                curve = PositionCurve(keys)
                output_animation.position_curves.append(
                    AnimationCurve(curve, joint_i)
                )
            elif target.path == "rotation":
                curve = QuaternionCurve(keys)
                output_animation.rotation_curves.append(
                    AnimationCurve(curve, joint_i)
                )

        output_animation.duration /= IFileLoader.DEFAULT_FRAMERATE

        return output_skeleton, output_animation

if __name__ == "__main__":
    glb = GlbLoader(Path("C:\dev\MotionMachine\skanym\src\skanym\examples\input\walk_glb.glb"))
    # skeleton = glb.load_skeleton()
    animation = glb.load_animation()
    for curve in animation.position_curves[0:3]:
        for key in curve.curve.keys:
            pass
            # print(key)


import numpy as np

from skanym.structures.data.iTransform import ITransform
from skanym.utils import conversion

class TransformMatrix(ITransform):
    """
    Concrete Transform class defined only by its transform matrix.
    """
    def __init__(self, transformMatrix: np.ndarray = np.identity(4)):
        """
        Constructor.
        """
        self.transformMatrix = transformMatrix

    def getPosition(self):
        """
        Returns the position of the transform.
        """
        return conversion.transformMatrixToPosition(self.transformMatrix)

    def getRotation(self):
        """
        Returns the rotation of the transform.
        """
        return conversion.rotationMatrixToQuaternion(self.getRotationMatrix())

    def getTransformMatrix(self):
        """
        Returns the transform matrix.
        """
        return self.transformMatrix

    def getRotationMatrix(self):
        """
        Returns the rotation matrix.
        """
        return self.transformMatrix[0:3, 0:3]

    def __str__(self):
        return "Position: " + str(self.getPosition()) + ", Rotation: " + str(self.getRotation())


from typing import List

import numpy as np

from skanym.structures.data.vectorizedArray import VectorizedArray
from skanym.structures.character.joint import Joint

class Skeleton:
    """A class to represent a skeleton (or rig).

    Attributes
    ----------
    name : str
        Name given to the skeleton. Not necessarily unique.
    root : Joint
        Root joint of the hierarchy for the skeleton.
    """

    def __init__(self, root: Joint, name: str = "new skeleton"):
        """**Default constructor for the Skeleton class.**

        Parameters
        ----------
        root : Joint
            Root joint of the hierarchy for the skeleton.
        name : str, optional
            Name of the skeleton, by default "new skeleton".
        """
        self.name = name
        self.root = root
   
    def as_joint_list(self) -> List[Joint]:
        """Returns the skeleton as a list of joints.

        Note
        ----
        The list is in depth-first order starting from the root.

        Returns
        -------
        (Joint) list
            List of joints in the skeleton.
        """
        adjacency_list = []
        stack = [self.root]
        while stack:
            joint = stack.pop()
            adjacency_list.append(joint)
            stack.extend(reversed(joint.children))
        return adjacency_list
    
    def as_joint_dict(self) -> dict:
        """Returns the skeleton as a dictionary of joints.

        Note
        ----
        The dictionary is in depth-first order starting from the root.

        Returns
        -------
        dict
            Dictionary of joints in the skeleton.
        """
        joint_dict = {}
        for joint in self.as_joint_list():
            joint_dict[self.get_joint_id(joint)] = joint.name
        return joint_dict

    def as_hierarchy_list(self) -> List[Joint]:
        """Returns the skeleton as a list of joints.

        Note
        ----
        The list is in breadth-first order starting from the root.

        Returns
        -------
        (Joint) list
            List of joints in the skeleton.
        """
        hierarchy_list = []
        current_level = [self.root]
        while current_level:
            hierarchy_list.append(current_level)
            next_level = []
            for joint in current_level:
                next_level.extend([child for child in joint.children])
            current_level = next_level
        return hierarchy_list  
    
    def as_bind_transform_vector(self) -> VectorizedArray:
        local_bind_matrices = []
        for joint in self.as_joint_list():
            local_bind_matrices.append(joint.local_bind_transform.getTransformMatrix())
        arr_local_bind_matrices = np.array(local_bind_matrices).reshape(-1, 4, 4)
        return VectorizedArray(
            array=arr_local_bind_matrices,
            labels=["local_bind_matrix"],
        )
    
    def as_parent_id_vector(self) -> VectorizedArray:
        parent_ids_list = [-(2 ** 31)]
        for joint in self.as_joint_list()[1:]:  # skip the root joint
            parent_id = self.get_joint_id(joint.parent)
            parent_ids_list.append(parent_id)
        return VectorizedArray(
            array=np.array(parent_ids_list, dtype=np.int32),
            labels=["parent_id"],
        )

    def get_joint_by_name(self, name):
        """Returns the joint with the given name.

        Parameters
        ----------
        name : str
            Name of the joint to return.

        Returns
        -------
        Joint
            Joint with the given name. None if not found.
        """
        for joint in self.depth_first_list:
            if joint.name == name:
                return joint
        return None

    def get_nb_joints(self):
        return len(self.as_joint_list())
    
    def get_joint_by_id(self, id):
        return self.as_joint_list()[id]
    
    def get_joint_id(self, joint: Joint):
        return self.as_joint_list().index(joint)
    

    def __repr__(self):
        return "Skeleton: " + self.name + ", Root: " + self.root.name
    


import numpy as np


class VectorizedArray:
    def __init__(self, array, labels=None):
        if not isinstance(array, np.ndarray):
            raise ValueError("Input must be a NumPy ndarray")
        self._array = array

        self._labels = labels
        if self._labels is None:
            if self._array.ndim == 1:
                self._labels = ["col_0"]
            else:
                self._labels = [f"col_{i}" for i in range(self._array.shape[1])]

    @classmethod
    def copy(cls, vectorized_array):
        return cls(
            array=np.copy(vectorized_array.get()),
            labels=vectorized_array.get_labels().copy(),
        )

    def get(self):
        return self._array

    def tile(self, n_tuple):
        tiled_vectorized_array = VectorizedArray.copy(self)
        tiled_vectorized_array._array = np.tile(self._array, n_tuple)
        return tiled_vectorized_array

    def repeat(self, n):
        repeated_vectorized_array = VectorizedArray.copy(self)
        repeated_vectorized_array._array = np.repeat(self._array, n, axis=0)
        return repeated_vectorized_array
    
    def pop(self, index):
        popped_value = self._array[index]
        self._array = np.delete(self._array, index, axis=0)
        return popped_value

    def get_labels(self):
        return self._labels

    def set_labels(self, labels):
        self._labels = labels

    def validate_shape(self, expected_shape):
        if self._array.shape != expected_shape:
            raise ValueError(
                f"Expected shape {expected_shape} but got {self._array.shape}"
            )

    @staticmethod  
    def create_transform_matrix_vector(N: int) -> "VectorizedArray":
        return VectorizedArray(
            array=np.tile(np.eye(4), (N, 1, 1)), labels=["transform_matrix"]
        )

    def __len__(self):
        return self._array.shape[0]

    def __str__(self):
        str_value = ""
        str_value += str(self._labels) + "\n"
        str_value += str(self._array) + "\n"
        str_value += str(self._array.shape) + "\n"
        return str_value

    def __getitem__(self, key):
        return self._array[key]

    def __setitem__(self, key, value):
        self._array[key] = value



from typing import List, Tuple
from enum import Enum

import numpy as np
import quaternion

from line_profiler import profile

from skanym.animators.baseAnimator import BaseAnimator
from skanym.structures.character.skeleton import Skeleton
from skanym.structures.animation.animation import Animation
from skanym.structures.data.vectorizedArray import VectorizedArray
from skanym.utils.conversion import quaternionToRotationMatrix

class PlayMode(Enum):
    LOOP_ALL = 0
    WAIT_THEN_LOOP = 1
    NO_LOOP = 2

class BaseFkAnimator(BaseAnimator):
    """ """

    def __init__(self, skeleton: Skeleton, animations: List[Animation], play_mode: PlayMode = PlayMode.LOOP_ALL):
        self.nb_anims = None
        self.nb_joints = None

        self.vec_durations = None
        self.vec_pos_curves = None
        self.vec_rot_curves = None
        self.vec_pos_curve_ids = None
        self.vec_rot_curve_ids = None

        self.vec_output = None

        self.play_mode = play_mode

        super().__init__(skeleton, animations)

    def _initialize(self):
        self.nb_anims = len(self.animations)
        self.nb_joints = self.skeleton.get_nb_joints()

        self.vec_output = VectorizedArray.create_transform_matrix_vector(
            self.nb_anims * self.nb_joints
        )

        self.vec_durations = Animation.animations_as_duration_vector(self.animations)
        self.vec_durations = self.vec_durations.repeat(self.nb_joints)

        lst_pos_curve_ids = []
        self.lst_vec_pos_curves = []
        for anim_id in range(self.nb_anims):
            for curve in self.animations[anim_id].position_curves:
                if curve.is_empty():
                    continue
                pos_curve_id = curve.id + anim_id * self.nb_joints
                if curve.is_constant():
                    self.vec_output[pos_curve_id][:3, 3] = curve.get_constant_value()
                else:
                    # 2 or more keys => requires interpolation
                    lst_pos_curve_ids.append(pos_curve_id)
                    self.lst_vec_pos_curves.append(curve.to_vectorized_array())

        self.vec_pos_curve_ids = VectorizedArray(
            array=np.array(lst_pos_curve_ids, dtype=np.int32),
            labels=["curve_id"],
        )

        lst_rot_curve_ids = []
        self.lst_vec_rot_curves = []
        for anim_id in range(self.nb_anims):
            for curve in self.animations[anim_id].rotation_curves:
                if curve.is_empty():
                    continue
                rot_curve_id = curve.id + anim_id * self.nb_joints
                if curve.is_constant():
                    self.vec_output[rot_curve_id][:3, :3] = quaternionToRotationMatrix(
                        quaternion.from_float_array(curve.get_constant_value())
                    )
                else:
                    # 2 or more keys => requires interpolation
                    lst_rot_curve_ids.append(rot_curve_id)
                    self.lst_vec_rot_curves.append(curve.to_vectorized_array())

        self.vec_rot_curve_ids = VectorizedArray(
            array=np.array(lst_rot_curve_ids, dtype=np.int32),
            labels=["curve_id"],
        )

        self.vec_pos_curves = self.vectorize_lst_curve_vectors(self.lst_vec_pos_curves)
        self.vec_rot_curves = self.vectorize_lst_curve_vectors(self.lst_vec_rot_curves)

    # @profile
    def _execute(self):
        lst_output = []
        nb_transforms = self.nb_joints * self.nb_anims

        vec_delta_transform_matrices = self.vec_output
        vec_delta_transform_matrices.set_labels(["delta_local_transform"])

        vec_anim_times = self.real_to_anim_time(self.time, self.vec_durations)

        arr_pos_addends = (
            np.arange(1).repeat(len(self.vec_pos_curve_ids)) * nb_transforms
        )
        arr_pos_curve_ids = self.vec_pos_curve_ids.get()
        arr_pos_curve_ids += arr_pos_addends

        arr_rot_addends = (
            np.arange(1).repeat(len(self.vec_rot_curve_ids)) * nb_transforms
        )
        arr_rot_curve_ids = self.vec_rot_curve_ids.get()
        arr_rot_curve_ids += arr_rot_addends
        arr_anim_times = vec_anim_times.get()
        vec_pos_anim_times = VectorizedArray(arr_anim_times[arr_pos_curve_ids])
        vec_rot_anim_times = VectorizedArray(arr_anim_times[arr_rot_curve_ids])

        # COMPUTATION
        vec_pos_prev_keys, vec_pos_next_keys = self.get_prev_next_key(
            vec_pos_anim_times, self.vec_pos_curves
        )

        vec_position_interpolation_ratios = self.compute_interpolation_ratio(
            vec_pos_anim_times, vec_pos_prev_keys, vec_pos_next_keys
        )
       
        vec_delta_position = self.interpolate_position_linear(
            vec_pos_prev_keys, vec_pos_next_keys, vec_position_interpolation_ratios
        )

        vec_rot_prev_keys, vec_rot_next_keys = self.get_prev_next_key(
            vec_rot_anim_times, self.vec_rot_curves
        )

        vec_rotation_interpolation_ratios = self.compute_interpolation_ratio(
            vec_rot_anim_times, vec_rot_prev_keys, vec_rot_next_keys
        )

        vec_delta_rotation = self.interpolate_quaternion_linear(
            vec_rot_prev_keys, vec_rot_next_keys, vec_rotation_interpolation_ratios
        )

        vec_delta_transform_matrices[arr_pos_curve_ids, :3, 3] = vec_delta_position

        vec_delta_transform_matrices[arr_rot_curve_ids, :3, :3] = (
            quaternionToRotationMatrix(vec_delta_rotation.get())
        )

        lst_output.append(vec_delta_transform_matrices)
        return lst_output

    @staticmethod
    def vectorize_lst_curve_vectors(
        curve_vectors: List[VectorizedArray],
    ) -> VectorizedArray:
        max_len = max([len(curve_vector) for curve_vector in curve_vectors])
        for curve_vector in curve_vectors:
            while len(curve_vector) < max_len:
                curve_vector._array = np.vstack(
                    [curve_vector._array, curve_vector._array[-1]]
                )

        return VectorizedArray(array=np.array(curve_vectors), labels=["curve_vectors"])

    # @profile
    def real_to_anim_time(
        self, real_time: float, vec_durations: VectorizedArray
    ) -> VectorizedArray:
            
        if self.play_mode == PlayMode.NO_LOOP:
            arr_anim_time = real_time / vec_durations.get()
            arr_anim_time = np.clip(arr_anim_time, 0, 1)
            vec_anim_time = VectorizedArray(array=arr_anim_time, labels=["anim_time"])
            return vec_anim_time
    
        elif self.play_mode == PlayMode.WAIT_THEN_LOOP:
            real_time = real_time % vec_durations.get().max()
            arr_anim_time = real_time / vec_durations.get()
            arr_anim_time = np.clip(arr_anim_time, 0, 1)
            if np.allclose(arr_anim_time, 1.0, atol=1e-5):
                arr_anim_time = real_time / vec_durations.get() % 1.0            
            vec_anim_time = VectorizedArray(array=arr_anim_time, labels=["anim_time"])
            return vec_anim_time
        
        elif self.play_mode == PlayMode.LOOP_ALL:
            arr_anim_time = real_time / vec_durations.get() % 1.0
            vec_anim_time = VectorizedArray(array=arr_anim_time, labels=["anim_time"])
            return vec_anim_time


    @staticmethod
    # @profile
    def get_prev_next_key(
        vec_anim_times: VectorizedArray, curve_vectors: VectorizedArray
    ) -> Tuple[VectorizedArray, VectorizedArray]:
        # Make a new 2D vectorized array that only contains the first column of each curve_vector
        arr_curve_times = curve_vectors[:, :, 0]

        # Reshape the anim_times to match the shape of vec_curve_times
        arr_anim_times = np.tile(
            vec_anim_times.get().reshape(-1, 1), (1, arr_curve_times.shape[1])
        )

        # Subtract the two arrays
        arr_diff = arr_curve_times - arr_anim_times

        # Add 2 to negative values
        arr_diff[arr_diff <= 0] += 2

        # Find the indices of the first positive value in each row
        arr_next_keys_id = np.argmin(arr_diff, axis=1)

        arr_prev_keys_id = arr_next_keys_id - 1

        prev_keys = curve_vectors[np.arange(len(arr_prev_keys_id)), arr_prev_keys_id]
        next_keys = curve_vectors[np.arange(len(arr_next_keys_id)), arr_next_keys_id]

        vec_prev_keys = VectorizedArray(array=prev_keys, labels=["prev_keys"])

        vec_next_keys = VectorizedArray(array=next_keys, labels=["next_keys"])

        return vec_prev_keys, vec_next_keys

    @staticmethod
    def compute_interpolation_ratio(
        vec_anim_times: VectorizedArray,
        vec_prev_keys: VectorizedArray,
        vec_next_keys: VectorizedArray,
    ) -> VectorizedArray:
        prev_key_times = vec_prev_keys[:, 0]
        next_key_times = vec_next_keys[:, 0]
        return VectorizedArray(
            array=(vec_anim_times.get() - prev_key_times)
            / (next_key_times - prev_key_times),
            labels=["interpolation_ratio"],
        )

    @staticmethod
    def interpolate_position_linear(
        vec_prev_keys: VectorizedArray,
        vec_next_keys: VectorizedArray,
        vec_interpolation_ratio: VectorizedArray,
    ) -> VectorizedArray:
        return VectorizedArray(
            array=vec_prev_keys.get()[:, 1:]
            + vec_interpolation_ratio.get().reshape(-1, 1)
            * (vec_next_keys.get()[:, 1:] - vec_prev_keys.get()[:, 1:]),
            labels=["delta_pos_x", "delta_pos_y", "delta_pos_z"],
        )

    @staticmethod
    def interpolate_quaternion_linear(
        vec_prev_keys: VectorizedArray,
        vec_next_keys: VectorizedArray,
        vec_interpolation_ratio: VectorizedArray,
    ) -> VectorizedArray:
        prev_values = quaternion.as_quat_array(vec_prev_keys.get()[:, 1:])
        next_values = quaternion.as_quat_array(vec_next_keys.get()[:, 1:])
        return VectorizedArray(
            array=quaternion.slerp(
                prev_values, next_values, 0, 1, vec_interpolation_ratio.get()
            ),
            labels=["delta_q_w", "delta_q_x", "delta_q_y", "delta_q_z"],
        )



from abc import ABC, abstractmethod
from typing import Dict, List

import numpy as np
from line_profiler import profile

from skanym.structures.character.skeleton import Skeleton
from skanym.structures.animation.animation import Animation
from skanym.structures.data.vectorizedArray import VectorizedArray

class BaseAnimator(ABC):
    """ """
    
    # Type alias for the output dictionary
    OutputDictType = Dict[str, Dict[str, List[float]]]
    # Description : Dict[joint_name, Dict[data_label, List[data_value]]]

    def __init__(self, skeleton: Skeleton, animations: List[Animation]):
        self.skeleton = skeleton
        self.animations = None
        self.time = 0.0

        self.set_animations(animations)

    def set_animations(self, animations: List[Animation]):
        self.animations = animations
        self._initialize()

    def set_time(self, time: float):
        self.time = time

    # @profile
    def step(self, delta_time: float = 0.0) -> List[VectorizedArray]:
        self.time += delta_time
        return self._execute()
    
    @abstractmethod
    def _initialize(self):
        pass

    @abstractmethod
    def _execute(self):
        pass

    # @profile
    def output_to_dict(self, lst_output_vector: List[VectorizedArray]) -> List[OutputDictType]:
        lst_output_dict = []
        lst_label = [output_vector.get_labels()[0] for output_vector in lst_output_vector]
        lst_joint_name = [joint.name for joint in self.skeleton.as_joint_list()]
        lst_output_list = [list(output_vector.get()) for output_vector in lst_output_vector]
        nb_output = len(lst_output_list)

        global_id = 0

        for a_id in range(self.nb_anims):
            output_dict = {}
            for j_id in range(self.nb_joints):
                j_name = lst_joint_name[j_id]
                if j_name not in output_dict.keys():
                    output_dict[j_name] = {}
                for l_id in range(nb_output):
                    label = lst_label[l_id]
                    output_dict[j_name][label] = list(lst_output_list[l_id][global_id])
                            
                global_id += 1

            lst_output_dict.append(output_dict)
            
        return lst_output_dict




import numpy as np

from skanym.core.animate.curve.iCurve import ICurve
  
class ScalarCurve(ICurve):
    """
    """
    def as_array(self) -> np.ndarray:
        """Returns the animation curve as a numpy array.

        Returns:
        ----------
        np.ndarray
            The animation curve as a numpy array.
        """
        if self.get_key_count() == 0:
            return np.array([
                [0.0, 0.0],
                [1.0, 1.0]])
        elif self.get_key_count() == 1:
            return np.array([
                [0.0, self.keys[0].value],
                [self.keys[0].time, self.keys[0].value],
                [1.0, self.keys[0].value]])
        else:
            return np.array([[key.time, key.value] for key in self.keys])


import sys
import warnings
from typing import List
from pathlib import Path

import numpy as np

from skanym.loaders.iFileLoader import IFileLoader
from skanym.structures.character.skeleton import Skeleton
from skanym.structures.character.joint import Joint
from skanym.structures.animation.animation import Animation
from skanym.structures.animation.animationCurve import AnimationCurve
from skanym.structures.animation.positionCurve import PositionCurve
from skanym.structures.animation.quaternionCurve import QuaternionCurve
from skanym.structures.animation.key import Key
from skanym.structures.data.transform import Transform


class FbxLoader(IFileLoader):
    def __init__(self, path: Path):
        super().__init__(path)
        try:
            import fbx
        except ImportError:
            import platform

            sys.path.append("C:/dev/MotionMachine/skanym/src/skanym/utils/fbx_bindings")

            # Second try:
            try:
                import fbx
            except ImportError:
                raise

        self._fbx = fbx

    def load_skeleton(self) -> Skeleton:
        fbx_scene = self._load_fbx_scene(self._path)

        root_node = fbx_scene.GetRootNode()
        skeleton_node_list = []

        self._load_skeleton_nodes(skeleton_node_list, root_node)

        skeleton_root = None

        joint_hierarchy = []

        for skeleton_node in skeleton_node_list:
            # Creating joints
            joint_name = skeleton_node.GetName()

            pos = np.array([float(val) for val in skeleton_node.LclTranslation.Get()])

            rotation = skeleton_node.PreRotation.Get()

            v = self._fbx.FbxVector4(rotation)

            m = self._fbx.FbxAMatrix()
            m.SetR(v)
            q = m.GetQ()
            orient = np.quaternion(q.GetAt(3), q.GetAt(0), q.GetAt(1), q.GetAt(2))
            # if orient.w < 0:
            #     print(f"NEGATIVE QUATERNION for {joint_name} in {fbx_file_path.split('/')[-1]}, {orient}")

            joint = Joint(name=joint_name, local_bind_transform=Transform(pos, orient))

            children = []
            for i in range(skeleton_node.GetChildCount()):
                child_node = skeleton_node.GetChild(i)
                children.append(child_node.GetName())

            joint_hierarchy.append((joint, children))

            if skeleton_root is None:
                skeleton_root = joint

        # Building joint hierarchy
        joint_list = [relation[0] for relation in joint_hierarchy]

        for relation in joint_hierarchy:
            children = [joint for joint in joint_list if joint.name in relation[1]]
            relation[0].add_children(children)

        # Creating skeleton
        skeleton_root.local_bind_transform = Transform()
        skeleton = Skeleton(skeleton_root)

        return skeleton

    def load_animation(self) -> Animation:
        fbx_scene = self._load_fbx_scene(self._path)

        root_node = fbx_scene.GetRootNode()
        skeleton_node_list = []

        self._load_skeleton_nodes(skeleton_node_list, root_node)

        anim_layer = self._get_anim_layer(fbx_scene)

        animation = Animation(
            name=self._path.stem,
            duration=0.0,
            shift=0.0,
            position_curves=[],
            rotation_curves=[],
        )

        node_id = 0
        for skeleton_node in skeleton_node_list:
            # Creating joint animations
            if anim_layer:
                translation_curve, rotation_curve, duration = self._load_curves(
                    skeleton_node, anim_layer
                )
                if duration > animation.duration:
                    animation.duration = duration
                animation.position_curves.append(
                    AnimationCurve(translation_curve, node_id)
                )
                animation.rotation_curves.append(
                    AnimationCurve(rotation_curve, node_id)
                )

            node_id += 1

        animation.duration /= IFileLoader.DEFAULT_FRAMERATE

        return animation

    def load_animations(self) -> List[Animation]:
        return NotImplementedError

    def _load_fbx_scene(self, fbx_file_path: Path):
        sdk_manager = self._fbx.FbxManager.Create()

        ios = self._fbx.FbxIOSettings.Create(sdk_manager, self._fbx.IOSROOT)
        sdk_manager.SetIOSettings(ios)

        importer = self._fbx.FbxImporter.Create(sdk_manager, "")

        if not importer.Initialize(str(fbx_file_path), -1, sdk_manager.GetIOSettings()):
            print("Call to FbxImporter::Initialize() failed.")
            raise ImportError(
                "Error importing file %s file not found or %s"
                % (
                    str(fbx_file_path),
                    importer.GetStatus().GetErrorString(),
                )
            )

        scene = self._fbx.FbxScene.Create(sdk_manager, "myScene")
        importer.Import(scene)
        importer.Destroy()

        return scene

    def _load_skeleton_nodes(self, skeleton_node_list, root_node):
        for i in range(root_node.GetChildCount()):
            child_node = root_node.GetChild(i)
            for i in range(child_node.GetNodeAttributeCount()):
                attribute = child_node.GetNodeAttributeByIndex(i)
                if type(attribute) == self._fbx.FbxSkeleton:
                    skeleton_node_list.append(child_node)
            self._load_skeleton_nodes(skeleton_node_list, child_node)

    def _get_anim_layer(self, scene):
        nb_anim_stack = scene.GetSrcObjectCount(
            self._fbx.FbxCriteria.ObjectType(self._fbx.FbxAnimStack.ClassId)
        )

        if nb_anim_stack == 0:
            raise ValueError("No animation stack found in fbx file.")
        elif nb_anim_stack > 1:
            warnings.warn(
                "Multiple anim stacks found in fbx file, only the first one is treated",
                stacklevel=2,
            )

        anim_stack = scene.GetSrcObject(
            self._fbx.FbxCriteria.ObjectType(self._fbx.FbxAnimStack.ClassId), 0
        )

        nb_anim_layers = anim_stack.GetSrcObjectCount(
            self._fbx.FbxCriteria.ObjectType(self._fbx.FbxAnimLayer.ClassId)
        )

        if nb_anim_layers == 0:
            raise ValueError("No animation layer found in anim stack.")
        elif nb_anim_layers > 1:
            warnings.warn(
                "Multiple anim layers found in anim stack, only the first one is treated",
                stacklevel=2,
            )

        anim_layer = anim_stack.GetSrcObject(
            self._fbx.FbxCriteria.ObjectType(self._fbx.FbxAnimLayer.ClassId), 0
        )

        return anim_layer

    def _load_curves(self, skeleton_node, anim_layer):
        # Assumes that when a change to the position/orientation of a joint is made,
        # the translation/rotation values for each axis are given in the keyframe.

        tX_curve = skeleton_node.LclTranslation.GetCurve(anim_layer, "X")
        tY_curve = skeleton_node.LclTranslation.GetCurve(anim_layer, "Y")
        tZ_curve = skeleton_node.LclTranslation.GetCurve(anim_layer, "Z")
        rX_curve = skeleton_node.LclRotation.GetCurve(anim_layer, "X")
        rY_curve = skeleton_node.LclRotation.GetCurve(anim_layer, "Y")
        rZ_curve = skeleton_node.LclRotation.GetCurve(anim_layer, "Z")

        translation_keys = []
        rotation_keys = []

        duration = 1.0

        if tX_curve is not None and tY_curve is not None and tZ_curve is not None:
            for key_id in range(tX_curve.KeyGetCount()):
                key_time = tX_curve.KeyGetTime(key_id).GetTimeString("")

                t = np.array(
                    [
                        tX_curve.KeyGetValue(key_id),
                        tY_curve.KeyGetValue(key_id),
                        tZ_curve.KeyGetValue(key_id),
                    ]
                )

                if "*" in key_time:
                    # Keys whose time is marked with "*" are not used in the animation.
                    # They are probably used to improve interpolation quality in between keyframes.
                    pass
                else:
                    key_time = float(key_time)
                    if key_time > duration:
                        duration = key_time

                    translation_keys.append(Key(key_time, t))

        if rX_curve is not None and rY_curve is not None and rZ_curve is not None:
            for key_id in range(rX_curve.KeyGetCount()):
                key_time = rX_curve.KeyGetTime(key_id).GetTimeString("")

                v = self._fbx.FbxVector4(
                    rX_curve.KeyGetValue(key_id),
                    rY_curve.KeyGetValue(key_id),
                    rZ_curve.KeyGetValue(key_id),
                )

                m = self._fbx.FbxAMatrix()
                m.SetR(v)
                q = m.GetQ()
                q = np.array([q.GetAt(3), q.GetAt(0), q.GetAt(1), q.GetAt(2)])

                # if q.w < 0:
                #     print(f"NEGATIVE QUATERNION for {skeleton_node.GetName()} in {fbx_file_path.split('/')[-1]} at time {key_time}, {q}")

                if "*" in key_time:
                    # Keys whose time is marked with "*" are not used in the animation.
                    # They are probably used to improve interpolation quality in between keyframes.
                    pass
                else:
                    key_time = float(key_time)
                    if key_time > duration:
                        duration = key_time

                    rotation_keys.append(Key(key_time, q))

        for key in translation_keys:
            key.time /= duration
        for key in rotation_keys:
            key.time /= duration

        translation_curve = PositionCurve(translation_keys)
        rotation_curve = QuaternionCurve(rotation_keys)
        return translation_curve, rotation_curve, duration

if __name__ == "__main__":
    fbx = FbxLoader(Path("C:\dev\MotionMachine\skanym\src\skanym\examples\input\walk.fbx"))
    skeleton = fbx.load_skeleton()
    for joint in skeleton.as_joint_list()[0:3]:
        print(type(joint.local_bind_transform.position))

    animation = fbx.load_animation()
    for curve in animation.position_curves[0:3]:
        for key in curve.curve.keys[0:3]:
            print(type(key.value))


from typing import List

import numpy as np

from line_profiler import profile

from skanym.structures.character.skeleton import Skeleton
from skanym.structures.animation.animation import Animation
from skanym.structures.data.vectorizedArray import VectorizedArray
from skanym.animators.baseFkAnimator import PlayMode
from skanym.animators.localFkAnimator import LocalFkAnimator

class GlobalFkAnimator(LocalFkAnimator):
    """ """

    def __init__(self, skeleton: Skeleton, animations: List[Animation], play_mode: PlayMode = PlayMode.LOOP_ALL):
        self.vec_parent_ids = None
        super().__init__(skeleton, animations, play_mode)

    # @profile
    def _initialize(self):
        super()._initialize()
        vec_parent_ids = self.skeleton.as_parent_id_vector()
        self.vec_parent_ids = vec_parent_ids.tile(self.nb_anims)

    @profile
    def _execute(self):
        lst_output = super()._execute()
        vec_local_transform_matrices = lst_output[1]

        arr_parent_ids = self.vec_parent_ids.tile(1)
        arr_parent_ids_addends = (
            np.arange(self.nb_anims).repeat(self.nb_joints) * self.nb_joints
        )
        arr_parent_ids += arr_parent_ids_addends
        vec_parent_ids = VectorizedArray(arr_parent_ids, labels=["parent_id"])

        arr_local_transform_matrices = vec_local_transform_matrices.get()
        arr_global_transform_matrices = np.copy(arr_local_transform_matrices)
        for i in range(self.nb_anims * self.nb_joints):
            parent_id = vec_parent_ids[i]
            if parent_id >= 0:
                arr_global_transform_matrices[i] = (
                    arr_global_transform_matrices[parent_id]
                    @ arr_local_transform_matrices[i]
                )

        vec_global_transform_matrices = VectorizedArray(
            arr_global_transform_matrices, labels=["global_transform"]
        )

        lst_output.append(vec_global_transform_matrices)
        return lst_output



import warnings
import copy
from typing import List

import numpy as np

from skanym.structures.data.iTransform import ITransform
from skanym.structures.data.transform import Transform

class Joint:
    """A class to represent a joint.

    A joint is considered a root if it has no parent.

    Attributes
    ----------
    id : int
        Unique id of the joint.
    name : str
        Name given to the joint. Not necessarily unique.
    parent : Joint
        Parent of the joint.
    children : (Joint) list
        List of children joints.
    local_bind_transform : Transform
        Initial transform of the joint in the parent's local coordinate system.
    """

    def __init__(self, name="new joint", parent=None, children = None, local_bind_transform: ITransform=Transform()):
        """**Default constructor for the Joint class.**

        Parameters
        ----------
        name : str, optional
            Name of the joint, by default "new joint".
        local_bind_transform : Transform, optional
            Initial transform relative to the parent, by default identity transform.
        """
        self.name = name
        self.parent: Joint = None
        self.children: List[Joint] = []
        self.local_bind_transform = local_bind_transform


    def add_child(self, joint: 'Joint'):
        """Makes the given joint a child of the current joint.

        Note
        ----
        The joint's parent is updated automatically.

        Parameters
        ----------
        joint : Joint
            The joint to be added as a child.

        Raises
        --------
        If the addition of the joint makes the tree cyclic, an error is raised.
        """
        if joint.parent is None:
            joint.parent = self
            self.children.append(joint)
        elif joint.parent != self:
            old_parent = joint.parent
            old_parent.children.remove(joint)
            joint.parent = self
            self.children.append(joint)
        if self.is_cyclic():
            warnings.warn(
                "Cyclic joint hierarchy detected. The add_child operation has been canceled.", UserWarning
            )
            self.remove_child(joint)

    def add_children(self, joints: List['Joint']):
        """Makes the given joints children of the current joint.

        Note
        ----
        The joints' parents are updated automatically.

        Parameters
        ----------
        joints : (Joint) list
            The joints to be added as children.
        """
        for joint in joints:
            self.add_child(joint)

    def remove_child(self, joint: 'Joint'):
        """Removes the given joint from the children of the current joint.

        Note
        ----
        If the joint is not a child of the current joint, then nothing happens.
        The joint's parent is updated automatically.

        Parameters
        ----------
        joint : Joint
            The joint to be removed.
        """
        if joint in self.children:
            self.children.remove(joint)
            joint.parent = None

    def remove_children(self, joints: List['Joint']):
        """Removes the given joints from the children of the current joint.

        Note
        ----
        The joints' parents are updated automatically.

        Parameters
        ----------
        joints : (Joint) list
            The joints to be removed.
        """
        for joint in joints:
            self.remove_child(joint)

    def remove_all_children(self):
        """Removes all the children of the current joint.

        Note
        ----
        The removed joints' parents are updated automatically.
        """
        self.remove_children(copy.copy(self.children))

    def is_cyclic(self):
        """Checks for cyclicity in the joint hierarchy for the current joint.

        Note: The joint hierarchy is considered cyclic only if a joint is the parent of one of its ancestors.

        Examples
        -------
        >>> root = Joint()
        >>> child1 = Joint()
        >>> child2 = Joint()
        >>> child3 = Joint()
        >>> root.add_children([child1, child2])
        >>> child1.add_child(child3)
        # The joint hierarchy is now:
        #   root -> child1 -> child3
        #        -> child2
        >>> root.is_cyclic()
        False
        >>> child3.add_child(child1)
        # The joint hierarchies are now:
        # root -> child2  and  ... -> child1 -> child3 -> child1 -> child3 -> child1 -> ...
        >>> root.is_cyclic()
        False
        >>> child1.is_cyclic()
        True

        Returns
        -------
        bool
            True if the joint hierarchy is cyclic, False otherwise.
        """

        # List of visited nodes.
        visited = []
        # Stack to keep track of the current node.
        stack = []
        stack.append(self)

        # Run till stack is empty or cycle is found.
        while stack:
            node = stack.pop()
            for child in node.children:
                if child in visited:
                    return True
                else:
                    stack.append(child)
                    visited.append(node)

        # If we reach here, then there is no cycle
        return False



import numpy as np
import quaternion

VALIDATION_DICT = {
    "positionToTransformMatrix": {
        "arg_name": "positions", "row_shape" : (3,), "dtype" : [np.float32, np.float64]},
    "transformMatrixToPosition": {
        "arg_name": "transform_matrices", "row_shape" : (4,4), "dtype" : [np.float32, np.float64]},
    "quaternionToRotationMatrix": {
        "arg_name": "quaternions", "row_shape" : (1,), "dtype" : np.quaternion},
    "rotationMatrixToQuaternion": {
        "arg_name": "rotation_matrices", "row_shape" : (3,3), "dtype" : [np.float32, np.float64]}    
    }

def safe(func):
    """
    Decorator method to check the validity of the input before conversion.
    """
    def wrapper(*args, **kwargs):
        # Get name of the function
        func_name = func.__name__
        # Get the input argument name
        arg_name = VALIDATION_DICT[func_name]["arg_name"]
        # Get the input argument value
        if args:
            arg_value = args[0]
        else:
            arg_value = kwargs[arg_name]
        # Get the row size of the input argument
        row_shape = VALIDATION_DICT[func_name]["row_shape"]
        # Get the data type of the input argument
        dtypes = VALIDATION_DICT[func_name]["dtype"]

        # Check the validity of the input argument
        if arg_value.ndim == 1:
            # If a single position is provided, convert it to a 2D array
            arg_value = np.expand_dims(arg_value, axis=0)
        if arg_value.shape[1:] != row_shape:
            expected_shape = (arg_value.shape[0],) + row_shape
            error_msg = f"Input {arg_name} array must have shape {row_shape} or {expected_shape} not {arg_value.shape}"
            raise ValueError(error_msg)
        if arg_value.dtype not in dtypes:
            error_msg = f"Input {arg_name} array must have data type {dtypes} not {arg_value.dtype}"
            raise ValueError(error_msg)  
             
        return func(*args, **kwargs)

    return wrapper

# @safe
def positionToTransformMatrix(positions):
    """
    Converts 3D positions transformation matrices.

    Parameters:
    - positions: NumPy array representing a single 3D position (shape: (3,))
                 or an array of 3D positions (shape: (n, 3))

    Returns:
    - transform_matrices: NumPy array representing transformation matrices
                          (shape: (n, 4, 4) for an array of positions, or (4, 4) for a single position)
    """
    if positions.ndim == 1:
        # If a single position is provided, convert it to a 2D array
        positions = np.expand_dims(positions, axis=0)

    # if positions.shape[1] != 3:
    #     raise ValueError("Input positions array must have shape (n, 3) or (3,)")

    # Create a 4x4 identity matrix for each position
    transform_matrices = np.tile(np.eye(4), (positions.shape[0], 1, 1))

    # Set the translation components in the transformation matrices
    transform_matrices[:, :3, 3] = positions

    return transform_matrices

# @safe
def transformMatrixToPosition(transform_matrices):
    """
    Converts transformation matrices to 3D positions.

    Parameters:
    - transform_matrices: NumPy array representing transformation matrices
                          (shape: (n, 4, 4))

    Returns:
    - positions: NumPy array representing 3D positions (shape: (n, 3))
    """
    # if transform_matrices.ndim != 3 or transform_matrices.shape[1:] != (4, 4):
    #     raise ValueError("Input transform_matrices array must have shape (n, 4, 4)")]
    
    return transform_matrices[:, :3, 3]

# @safe
def quaternionToRotationMatrix(quaternions):
    """
    Converts quaternions to rotation matrices.

    Parameters:
    - quaternions: NumPy array representing quaternions (shape: (n, 4))
    
    Returns:
    - rotation_matrices: NumPy array representing rotation matrices (shape: (n, 3, 3))
    """
    if quaternions.ndim == 0:
        # If a single quaternion is provided, convert it to a 2D array
        quaternions = np.expand_dims(quaternions, axis=0)
    
    # # Make a decorator class for validation
    # if quaternions.ndim != 2 or quaternions.shape[1] != 1:
    #     raise ValueError("Input quaternions array must have shape (n, 1)")
        
    # Assumes the quaternions are normalized
    m = np.empty(quaternions.shape + (3, 3))
    q = quaternion.as_float_array(quaternions)
    m[..., 0, 0] = 1.0 - 2*(q[..., 2]**2 + q[..., 3]**2)
    m[..., 0, 1] = 2*(q[..., 1]*q[..., 2] - q[..., 3]*q[..., 0])
    m[..., 0, 2] = 2*(q[..., 1]*q[..., 3] + q[..., 2]*q[..., 0])
    m[..., 1, 0] = 2*(q[..., 1]*q[..., 2] + q[..., 3]*q[..., 0])
    m[..., 1, 1] = 1.0 - 2*(q[..., 1]**2 + q[..., 3]**2)
    m[..., 1, 2] = 2*(q[..., 2]*q[..., 3] - q[..., 1]*q[..., 0])
    m[..., 2, 0] = 2*(q[..., 1]*q[..., 3] - q[..., 2]*q[..., 0])
    m[..., 2, 1] = 2*(q[..., 2]*q[..., 3] + q[..., 1]*q[..., 0])
    m[..., 2, 2] = 1.0 - 2*(q[..., 1]**2 + q[..., 2]**2)

    return m

# @safe
def rotationMatrixToQuaternion(rotation_matrices):
    """
    Converts rotation matrices to quaternions.

    Parameters:
    - rotation_matrices: NumPy array representing rotation matrices (shape: (n, 3, 3))

    Returns:
    - quaternions: NumPy array representing quaternions (shape: (n, 4))
    """
    if rotation_matrices.ndim == 2:
        # If a single rotation matrix is provided, convert it to a 3D array
        rotation_matrices = np.expand_dims(rotation_matrices, axis=0)
    
    # if rotation_matrices.ndim != 3 or rotation_matrices.shape[1:] != (3, 3):
    #     raise ValueError("Input rotation_matrices array must have shape (n, 3, 3)")
        
    rot = np.array(rotation_matrices, copy=False)
    shape = rot.shape[:-2]

    diagonals = np.empty(shape+(4,))
    diagonals[..., 0] = rot[..., 0, 0]
    diagonals[..., 1] = rot[..., 1, 1]
    diagonals[..., 2] = rot[..., 2, 2]
    diagonals[..., 3] = rot[..., 0, 0] + rot[..., 1, 1] + rot[..., 2, 2]

    indices = np.argmax(diagonals, axis=-1)

    q = diagonals  # reuse storage space
    indices_i = (indices == 0)
    if np.any(indices_i):
        if indices_i.shape == ():
            indices_i = Ellipsis
        rot_i = rot[indices_i, :, :]
        q[indices_i, 0] = rot_i[..., 2, 1] - rot_i[..., 1, 2]
        q[indices_i, 1] = 1 + rot_i[..., 0, 0] - rot_i[..., 1, 1] - rot_i[..., 2, 2]
        q[indices_i, 2] = rot_i[..., 0, 1] + rot_i[..., 1, 0]
        q[indices_i, 3] = rot_i[..., 0, 2] + rot_i[..., 2, 0]
    indices_i = (indices == 1)
    if np.any(indices_i):
        if indices_i.shape == ():
            indices_i = Ellipsis
        rot_i = rot[indices_i, :, :]
        q[indices_i, 0] = rot_i[..., 0, 2] - rot_i[..., 2, 0]
        q[indices_i, 1] = rot_i[..., 1, 0] + rot_i[..., 0, 1]
        q[indices_i, 2] = 1 - rot_i[..., 0, 0] + rot_i[..., 1, 1] - rot_i[..., 2, 2]
        q[indices_i, 3] = rot_i[..., 1, 2] + rot_i[..., 2, 1]
    indices_i = (indices == 2)
    if np.any(indices_i):
        if indices_i.shape == ():
            indices_i = Ellipsis
        rot_i = rot[indices_i, :, :]
        q[indices_i, 0] = rot_i[..., 1, 0] - rot_i[..., 0, 1]
        q[indices_i, 1] = rot_i[..., 2, 0] + rot_i[..., 0, 2]
        q[indices_i, 2] = rot_i[..., 2, 1] + rot_i[..., 1, 2]
        q[indices_i, 3] = 1 - rot_i[..., 0, 0] - rot_i[..., 1, 1] + rot_i[..., 2, 2]
    indices_i = (indices == 3)
    if np.any(indices_i):
        if indices_i.shape == ():
            indices_i = Ellipsis
        rot_i = rot[indices_i, :, :]
        q[indices_i, 0] = 1 + rot_i[..., 0, 0] + rot_i[..., 1, 1] + rot_i[..., 2, 2]
        q[indices_i, 1] = rot_i[..., 2, 1] - rot_i[..., 1, 2]
        q[indices_i, 2] = rot_i[..., 0, 2] - rot_i[..., 2, 0]
        q[indices_i, 3] = rot_i[..., 1, 0] - rot_i[..., 0, 1]

    q /= np.linalg.norm(q, axis=-1)[..., np.newaxis]

    return quaternion.as_quat_array(q)

def quaternionFromAxisAngle(axis, angle):
    """
    Creates a quaternion from an axis and an angle.

    Parameters:
    - axis: NumPy array representing the axis (shape: (3,))
    - angle: Angle in radians

    Returns:
    - quaternion: NumPy array representing the quaternion (shape: (4,))
    """
    # Normalize the axis
    axis = axis / np.linalg.norm(axis)

    # Create the quaternion
    half_angle = angle / 2
    return np.quaternion(np.cos(half_angle), *np.sin(half_angle) * axis)

def positionAndQuaternionToTransformMatrix(positions, quaternions):
    """
    Converts 3D positions and quaternions to transformation matrices.

    Parameters:
    - positions: NumPy array representing 3D positions (shape: (n, 3))
    - quaternions: NumPy array representing quaternions (shape: (n, 4))

    Returns:
    - transform_matrices: NumPy array representing transformation matrices (shape: (n, 4, 4))
    """
    # Convert positions and quaternions to transformation matrices
    transform_matrices = positionToTransformMatrix(positions)
    rotation_matrices = quaternionToRotationMatrix(quaternions)
    transform_matrices[:, :3, :3] = rotation_matrices

    return transform_matrices

if __name__ == "__main__":
    # Test the conversion functions

    # Single input
    print("____Single input")

    position = np.array([-5., 0., 3.])
    transform_matrix = positionToTransformMatrix(position)
    print("Transform matrix:", transform_matrix)

    rotation = quaternion.quaternion(np.sqrt(2)/2, 0, 0, np.sqrt(2)/2)
    rotation_matrix = quaternionToRotationMatrix(rotation)
    print("Rotation matrix:", rotation_matrix)

    transform_matrix = np.eye(4)
    transform_matrix[:3, 3] = position
    transform_matrix[:3, :3] = rotation_matrix

    position = transformMatrixToPosition(transform_matrix)
    print("Position:", position)

    rotation = rotationMatrixToQuaternion(rotation_matrix)
    print("Rotation:", rotation)

    # Multiple inputs
    print("____Multiple input")
    
    positions = np.array([[-5., 0., 3.], [0., 1., 2.]])
    transform_matrices = positionToTransformMatrix(positions)
    print("Transform matrices:", transform_matrices)

    rotations = np.array(
        [np.quaternion(np.sqrt(2)/2, 0, 0, np.sqrt(2)/2),
            np.quaternion(0, np.sqrt(2)/2, 0, np.sqrt(2)/2)],
        dtype=np.quaternion)
    rotation_matrices = quaternionToRotationMatrix(rotations)
    print("Rotation matrices:", rotation_matrices)

    transform_matrices = np.empty((2, 4, 4))
    transform_matrices[:, :3, 3] = positions
    transform_matrices[:, :3, :3] = rotation_matrices

    positions = transformMatrixToPosition(transform_matrices)
    print("Positions:", positions)

    rotations = rotationMatrixToQuaternion(rotation_matrices)
    print("Rotations:", rotations)




from fbx import *
import sys

def InitializeSdkObjects():
    # The first thing to do is to create the FBX SDK manager which is the 
    # object allocator for almost all the classes in the SDK.
    lSdkManager = FbxManager.Create()
    if not lSdkManager:
        sys.exit(0)
        
    # Create an IOSettings object
    ios = FbxIOSettings.Create(lSdkManager, IOSROOT)
    lSdkManager.SetIOSettings(ios)
    
    # Create the entity that will hold the scene.
    lScene = FbxScene.Create(lSdkManager, "")
    
    return (lSdkManager, lScene)

def SaveScene(pSdkManager, pScene, pFilename, pFileFormat = -1, pEmbedMedia = False):
    lExporter = FbxExporter.Create(pSdkManager, "")
    if pFileFormat < 0 or pFileFormat >= pSdkManager.GetIOPluginRegistry().GetWriterFormatCount():
        pFileFormat = pSdkManager.GetIOPluginRegistry().GetNativeWriterFormat()
        if not pEmbedMedia:
            lFormatCount = pSdkManager.GetIOPluginRegistry().GetWriterFormatCount()
            for lFormatIndex in range(lFormatCount):
                if pSdkManager.GetIOPluginRegistry().WriterIsFBX(lFormatIndex):
                    lDesc = pSdkManager.GetIOPluginRegistry().GetWriterFormatDescription(lFormatIndex)
                    if "ascii" in lDesc:
                        pFileFormat = lFormatIndex
                        break
    
    if not pSdkManager.GetIOSettings():
        ios = FbxIOSettings.Create(pSdkManager, IOSROOT)
        pSdkManager.SetIOSettings(ios)
    
    pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_MATERIAL, True)
    pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_TEXTURE, True)
    pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_EMBEDDED, pEmbedMedia)
    pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_SHAPE, True)
    pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_GOBO, True)
    pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_ANIMATION, True)
    pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_GLOBAL_SETTINGS, True)

    result = lExporter.Initialize(pFilename, pFileFormat, pSdkManager.GetIOSettings())
    if result == True:
        result = lExporter.Export(pScene)

    lExporter.Destroy()
    return result
    
def LoadScene(pSdkManager, pScene, pFileName):
    lImporter = FbxImporter.Create(pSdkManager, "")    
    result = lImporter.Initialize(pFileName, -1, pSdkManager.GetIOSettings())
    if not result:
        return False
    
    if lImporter.IsFBX():
        pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_MATERIAL, True)
        pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_TEXTURE, True)
        pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_EMBEDDED, True)
        pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_SHAPE, True)
        pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_GOBO, True)
        pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_ANIMATION, True)
        pSdkManager.GetIOSettings().SetBoolProp(EXP_FBX_GLOBAL_SETTINGS, True)
    
    result = lImporter.Import(pScene)
    lImporter.Destroy()
    return result



import numpy as np

class Key:

    def __init__(self, time, value):
        """**Default constructor for the Key class.**

        Parameters:
        ----------
        time : int or float
            The time of the key.
        value : any
            The values of the key at a given time.
        """
        self.time = time
        self.value = value

    def as_array(self) -> np.ndarray:
        """Returns the key as a numpy array.

        Returns:
        ----------
        np.ndarray
            The key as a numpy array.
        """
        return np.array([self.time, *self.value])
        
    def __repr__(self) -> str:
        """Returns the string representation of the key.

        Returns:
        ----------
        str
            The string representation of the key.
        """
        return f"Key({self.time}, {self.value})"


import pickle
from pathlib import Path

class ObjectLoader:
    def __init__(self, path: Path):
        self._path = path

    def set_path(self, path: Path):
        self._path = path

    def load(self) -> object:
        with open(self._path, "rb") as f:
            return pickle.load(f)
        
    def save(self, obj: object):
        with open(self._path, "wb") as f:
            pickle.dump(obj, f)


from abc import ABC, abstractmethod

import numpy as np
import quaternion

class ITransform(ABC):
    """
    Interface for transform objects.
    """

    @abstractmethod
    def getPosition(self) -> np.ndarray:
        """
        Returns the position of the transform.
        """
        pass

    @abstractmethod
    def getRotation(self) -> np.quaternion:
        """
        Returns the rotation of the transform.
        """
        pass

    @abstractmethod
    def getTransformMatrix(self) -> np.ndarray:
        """
        Returns the transform matrix.
        """
        pass
    
    @abstractmethod
    def getRotationMatrix(self) -> np.ndarray:
        """
        Returns the rotation matrix.
        """
        pass



import numpy as np

from skanym.structures.animation.iCurve import ICurve

class AnimationCurve:
    def __init__(self, curve: ICurve = None, id: int = None):
        self.curve = curve
        self.id = id

    def is_empty(self):
        return self.curve.is_empty()
    
    def is_constant(self):
        return self.curve.is_constant()
    
    def get_constant_value(self):
        return self.curve.get_constant_value()
    
    def to_vectorized_array(self):
        return self.curve.to_vectorized_array()

    





from typing import List
from pathlib import Path
from abc import ABC, abstractmethod

from skanym.structures.character.skeleton  import Skeleton
from skanym.structures.animation.animation  import Animation

class IFileLoader(ABC):

    DEFAULT_FRAMERATE = 30

    def __init__(self, path: Path):
        self._path = path

    def set_path(self, path: Path):
        self._path = path

    @abstractmethod
    def load_skeleton(self) -> Skeleton:
        pass

    @abstractmethod
    def load_animation(self) -> Animation:
        pass

    @abstractmethod
    def load_animations(self) -> List[Animation]:
        pass





from typing import List

import numpy as np
import quaternion
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler, MinMaxScaler

from line_profiler import profile

from skanym.structures.character.skeleton import Skeleton
from skanym.structures.animation.animation import Animation
from skanym.structures.animation.animationCurve import AnimationCurve
from skanym.structures.animation.positionCurve import PositionCurve
from skanym.structures.animation.quaternionCurve import QuaternionCurve
from skanym.structures.animation.key import Key
from skanym.structures.data.vectorizedArray import VectorizedArray
from skanym.animators.baseFkAnimator import BaseFkAnimator
from skanym.utils import conversion


class PcaAnimator(BaseFkAnimator):
    """ """

    def __init__(
        self,
        skeleton: Skeleton,
        animations: List[Animation],
        nb_components: int,
        nb_frames: int,
    ):
        self.nb_components = nb_components
        self.nb_frames = nb_frames
        self.pca_values = np.full(self.nb_components, 0.5)
        self.sampling_times = np.linspace(0, 1, self.nb_frames)

        self.vec_bind_transform_matrices = None
        self.vec_parent_ids = None
        super().__init__(skeleton, animations)

    def set_pca_values(self, pca_values: np.ndarray):
        self.pca_values = pca_values

    # @profile
    def _initialize(self):
        for anim in self.animations:
            anim.duration = 1.0
        super()._initialize()
        self.vec_bind_transform_matrices = self.skeleton.as_bind_transform_vector()

        # Input data generation
        real_time = self.time # To avoid losing the original time if set before _initialize

        vec_delta_transform_matrices = VectorizedArray.create_transform_matrix_vector(
            self.nb_joints * self.nb_anims * self.nb_frames
        )
        for sample_time_id in range(self.nb_frames):
            sample_time = self.sampling_times[sample_time_id]
            self.set_time(sample_time)
            start_idx = sample_time_id * self.nb_joints * self.nb_anims
            end_idx = (sample_time_id + 1) * self.nb_joints * self.nb_anims
            vec_delta_transform_matrices[start_idx:end_idx, :, :] = super()._execute()[0]
        
        self.set_time(real_time)

        arr_delta_positions = vec_delta_transform_matrices[:, :3, 3]
        arr_delta_rotations = quaternion.as_float_array(
            conversion.rotationMatrixToQuaternion(
                vec_delta_transform_matrices[:, :3, :3]
            )
        )

        neg_arr_delta_rotations = arr_delta_rotations[:, 0] < 0
        arr_delta_rotations[neg_arr_delta_rotations] *= -1
        input_data = np.concatenate((arr_delta_positions, arr_delta_rotations), axis=1)

        nb_pose = self.nb_joints * self.nb_anims

        indices = (
            np.arange(nb_pose * self.nb_frames)
            .reshape(self.nb_frames, nb_pose)
            .T.ravel()
        )
        input_data = input_data[indices]

        input_data = input_data.reshape((self.nb_anims, -1, input_data.shape[1]))
        input_data = input_data.reshape((-1, input_data.shape[1] * input_data.shape[2]))
        # PCA
        self.input_scaler = StandardScaler()
        self.output_scaler = MinMaxScaler()
        self.model = PCA(n_components=self.nb_components)

        scaled_input_data = self.input_scaler.fit_transform(input_data)
        output_data = self.model.fit_transform(scaled_input_data)
        self.scaled_output_data = self.output_scaler.fit_transform(output_data)

        self.vec_parent_ids = self.skeleton.as_parent_id_vector()

        self.nb_anims = 1

    # @profile
    def _execute(self):
        rescaled_values = self.output_scaler.inverse_transform(
            self.pca_values.reshape(1, -1)
        )
        generated_input = self.model.inverse_transform(rescaled_values)
        rescaled_input = self.input_scaler.inverse_transform(generated_input)   

        arr_pos_x = rescaled_input[:, 0::7]
        arr_pos_y = rescaled_input[:, 1::7]
        arr_pos_z = rescaled_input[:, 2::7]

        arr_rot_w = rescaled_input[:, 3::7]
        arr_rot_x = rescaled_input[:, 4::7]
        arr_rot_y = rescaled_input[:, 5::7]
        arr_rot_z = rescaled_input[:, 6::7]

        arr_pos = np.stack((arr_pos_x, arr_pos_y, arr_pos_z), axis=2)
        arr_rot = np.stack((arr_rot_w, arr_rot_x, arr_rot_y, arr_rot_z), axis=2)

        arr_pos = arr_pos.reshape(arr_pos.shape[1], arr_pos.shape[2])
        arr_rot = arr_rot.reshape(arr_rot.shape[1], arr_rot.shape[2])
        
        # Re normalize quaternions
        norms = np.linalg.norm(arr_rot, axis=1)
        arr_rot = arr_rot / norms[:, np.newaxis]

        arr_key_times = np.tile(self.sampling_times, self.nb_joints)
        arr_pos_key = np.concatenate(
            (arr_key_times.reshape(-1, 1), arr_pos), axis=1
        )
        arr_rot_key = np.concatenate(
            (arr_key_times.reshape(-1, 1), arr_rot), axis=1
        )

        arr_pos_key = arr_pos_key.reshape(self.nb_joints, self.nb_frames, 4)
        arr_rot_key = arr_rot_key.reshape(self.nb_joints, self.nb_frames, 5)

        vec_pos_curve = VectorizedArray(arr_pos_key, labels=["pos_curve"])
        vec_rot_curve = VectorizedArray(arr_rot_key, labels=["rot_curve"])

        vec_anim_times = self.real_to_anim_time(self.time, self.vec_durations)
        vec_anim_times._array = vec_anim_times[0:self.nb_joints]

        # COMPUTATION
        vec_pos_prev_keys, vec_pos_next_keys = self.get_prev_next_key(
            vec_anim_times, vec_pos_curve
        )

        vec_position_interpolation_ratios = self.compute_interpolation_ratio(
            vec_anim_times, vec_pos_prev_keys, vec_pos_next_keys
        )

        vec_delta_position = self.interpolate_position_linear(
            vec_pos_prev_keys, vec_pos_next_keys, vec_position_interpolation_ratios
        )

        vec_rot_prev_keys, vec_rot_next_keys = self.get_prev_next_key(
            vec_anim_times, vec_rot_curve
        )

        vec_rotation_interpolation_ratios = self.compute_interpolation_ratio(
            vec_anim_times, vec_rot_prev_keys, vec_rot_next_keys
        )

        vec_delta_rotation = self.interpolate_quaternion_linear(
            vec_rot_prev_keys, vec_rot_next_keys, vec_rotation_interpolation_ratios
        )

        vec_delta_transform_matrices = VectorizedArray.create_transform_matrix_vector(
           1 * self.nb_joints
        )

        vec_delta_transform_matrices[:, :3, 3] = vec_delta_position

        vec_delta_transform_matrices[:, :3, :3] = conversion.quaternionToRotationMatrix(
            vec_delta_rotation.get()
        )

        lst_output = [vec_delta_transform_matrices]

        arr_local_transform_matrices = self.vec_bind_transform_matrices.get() @ vec_delta_transform_matrices.get()

        vec_local_transform_matrices = VectorizedArray(arr_local_transform_matrices, labels=["local_transform"])

        lst_output.append(vec_local_transform_matrices)

        arr_local_transform_matrices = vec_local_transform_matrices.get()
        arr_global_transform_matrices = np.copy(arr_local_transform_matrices)
        for i in range(self.nb_joints):
            parent_id = self.vec_parent_ids[i]
            if parent_id >= 0:
                arr_global_transform_matrices[i] = (
                    arr_global_transform_matrices[parent_id]
                    @ arr_local_transform_matrices[i]
                )

        vec_global_transform_matrices = VectorizedArray(
            arr_global_transform_matrices, labels=["global_transform"]
        )
        lst_output.append(vec_global_transform_matrices)

        return lst_output

    # @profile
    def anim_from_pca(self, rescaled_input: np.ndarray) -> Animation:

        arr_pos_x = rescaled_input[:, 0::7]
        arr_pos_y = rescaled_input[:, 1::7]
        arr_pos_z = rescaled_input[:, 2::7]

        arr_rot_w = rescaled_input[:, 3::7]
        arr_rot_x = rescaled_input[:, 4::7]
        arr_rot_y = rescaled_input[:, 5::7]
        arr_rot_z = rescaled_input[:, 6::7]

        arr_pos = np.stack((arr_pos_x, arr_pos_y, arr_pos_z), axis=2)
        arr_rot = np.stack((arr_rot_w, arr_rot_x, arr_rot_y, arr_rot_z), axis=2)

        arr_pos = arr_pos.reshape(arr_pos.shape[1], arr_pos.shape[2])
        arr_rot = arr_rot.reshape(arr_rot.shape[1], arr_rot.shape[2])

        p_anim_curves = []
        r_anim_curves = []

        idx = 0
        for j_idx in range(self.nb_joints):
            p_keys = []
            r_keys = []
            for t_idx in range(self.nb_frames):
                p_keys.append(Key(self.sampling_times[t_idx], arr_pos[idx]))
                r_keys.append(Key(self.sampling_times[t_idx], arr_rot[idx]))
                idx += 1

            p_curve = PositionCurve(p_keys)
            r_curve = QuaternionCurve(r_keys)
            p_anim_curves.append(AnimationCurve(p_curve, j_idx))
            r_anim_curves.append(AnimationCurve(r_curve, j_idx))

        return Animation(position_curves=p_anim_curves, rotation_curves=r_anim_curves)


from typing import List
from abc import ABC, abstractmethod

import numpy as np

from skanym.structures.animation.key import Key

class ICurve(ABC):
    """An interface for curve classes."""
    def __init__(self, keys: List[Key] = None):
        """**Default constructor for the PositionCurve class.**"""
        self.keys = []
        if keys is not None:
            for key in keys:
                self.add_key(key)
            
    def get_key_count(self) -> int:
        """Returns the number of keys in the animation curve.

        Returns:
        ----------
        int
            The number of keys in the animation curve.
        """
        return len(self.keys)

    # @profile
    def add_key(self, key: Key):       
        if not 0.0 <= key.time <= 1.0:
            raise ValueError(
                f"Invalid key time {key.time}. Key time must be between 0.0 and 1.0 inclusive."
            )
        
        # self.validate_key_value(key.value)  

        self.keys.append(key)
        self.keys.sort(key=lambda key: key.time)

    def is_empty(self) -> bool:
        """Returns whether the animation curve is empty.

        Returns:
        ----------
        bool
            Whether the animation curve is empty.
        """
        return len(self.keys) == 0
    
    def is_constant(self) -> bool:
        """Returns whether the animation curve is constant.

        Returns:
        ----------
        bool
            Whether the animation curve is constant.
        """
        return len(self.keys) == 1
    
    def get_constant_value(self) -> Key:
        """Returns the constant key of the animation curve.

        Returns:
        ----------
        Key
            The constant key of the animation curve.
        """
        if not self.is_constant():
            raise ValueError("The animation curve is not constant.")
        return self.keys[0].value

    @abstractmethod
    def as_array(self) -> np.ndarray:
        pass




import numpy as np
import quaternion

from skanym.structures.data.iTransform import ITransform
from skanym.utils import conversion

class Transform(ITransform):
    """
    Simple concrete transform class defined by a position and a rotation.
    """
    def __init__(
            self,
            position: np.ndarray = np.array([0.0, 0.0, 0.0]),
            rotation: np.quaternion = quaternion.one):
        """
        Constructor.
        """
        self.position = position
        self.rotation = rotation

    def getPosition(self):
        """
        Returns the position of the transform.
        """
        return self.position

    def getRotation(self):
        """
        Returns the rotation of the transform.
        """
        return self.rotation

    def getTransformMatrix(self):
        """
        Returns the transform matrix.
        """        
        return conversion.positionAndQuaternionToTransformMatrix(self.position, self.rotation)

    def getRotationMatrix(self):
        """
        Returns the rotation matrix.
        """
        return conversion.quaternionToRotationMatrix(self.rotation)

    def __str__(self):
        return "Position: " + str(self.getPosition()) + ", Rotation: " + str(self.getRotation())
    


import numpy as np

from typing import List

from skanym.structures.data.vectorizedArray import VectorizedArray
from skanym.structures.animation.animationCurve import AnimationCurve

class Animation:
    def __init__(self,
                 name: str = "new_animation",
                 duration: float = 1.0,
                 shift: float = 0.0,
                 position_curves: List[AnimationCurve] = None,
                 rotation_curves: List[AnimationCurve] = None):
        self.name = name
        self.duration = duration
        self.shift = shift
        if position_curves is None:
            self.position_curves = []
        else:
            self.position_curves = position_curves

        if rotation_curves is None:
            self.rotation_curves = []
        else:
            self.rotation_curves = rotation_curves

    def get_position_curves_joint_ids(self) -> List[int]:
        return [curve.joint_id for curve in self.position_curves]
    
    def get_rotation_curves_joint_ids(self) -> List[int]:
        return [curve.joint_id for curve in self.rotation_curves]
        
    def get_position_curves_count(self) -> int:
        return len(self.position_curves)
    
    def get_rotation_curves_count(self) -> int:
        return len(self.rotation_curves)
    
    @staticmethod
    def animations_as_duration_vector(animations:List["Animation"]) -> VectorizedArray:
        lst_durations = [anim.duration for anim in animations]
        arr_durations = np.array(lst_durations)
        return VectorizedArray(array=arr_durations, labels=["duration"])
    
        

    


import numpy as np

from skanym.structures.animation.iCurve import ICurve  
from skanym.structures.data.vectorizedArray import VectorizedArray 
  
class QuaternionCurve(ICurve):
    """
    """
    def to_vectorized_array(self):
        lst = [key.as_array() for key in self.keys]
        arr = np.array(lst)
        return VectorizedArray(array=arr, labels=["time", "w", "x", "y", "z"])

    def as_array(self) -> np.ndarray:
        """Returns the animation curve as a numpy array.

        Returns:
        ----------
        np.ndarray
            The animation curve as a numpy array.
        """
        if self.get_key_count() == 0:
            return np.array([
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 1.0, 0.0, 0.0, 0.0]])
        elif self.get_key_count() == 1:
            return np.array([
                [0.0, self.keys[0].value.w, self.keys[0].value.x, self.keys[0].value.y, self.keys[0].value.z],
                [1.0, self.keys[0].value.w, self.keys[0].value.x, self.keys[0].value.y, self.keys[0].value.z]])
        else:
            return np.array([[key.time, key.value.w, key.value.x, key.value.y, key.value.z] for key in self.keys])